// rearranged LD3320 sample code, for tests on both asr and mp3 playing by ken woo 20220614.v.0.2.
// 20220618 v.0.3, support esp-m3 which disabled tx/rx. and with some structural modifies.


#include "LD3320.h"


// esp8266 speed up.
#if ESP8266_SPEEDUP
    IRAM_ATTR volatile unsigned &PIN_OUT_SET = *((unsigned*)0x60000304);
    IRAM_ATTR volatile unsigned &PIN_OUT_CLEAR = *((unsigned*)0x60000308);
    IRAM_ATTR volatile unsigned &PIN_IN = *((unsigned*)0x60000318);
#endif

// asr.
InTheIram unsigned nAsrStatus=LD_ASR_IDLE;              // interrupt result state.

InTheIram unsigned LD3320::nLD_Mode=LD_MODE_IDLE;       // the mode state.
InTheIram unsigned LD3320::nLD_reset=1;                 // demand to reset/go for whole procedure.
InTheIram unsigned LD3320::nLD_change_table=0;          // demand to change table.
InTheIram char *(*LD3320::var_tbl)[];                   // the table pointer used for loading the table.
IRAM_ATTR LD3320::p1fptr LD3320::ld_fptr=0;             // for asr. callback if set, and used in one-shot.

// mp3.
InTheIram File LD3320MP3::fh=File(0);
InTheIram unsigned LD3320MP3::flock=0;
InTheIram const char *LD3320MP3::var_mp3_filename=0;    // set a mp3 file name for playing.


/************************************************************************
功能描述：    io functions
入口参数：  none
返 回 值：   none
其他说明：  none
**************************************************************************/
void LD3320::LD_Init(){
    pinMode(SDCK, OUTPUT);
    pinMode(SDI, OUTPUT);
    pinMode(SCS, OUTPUT);
    pinMode(RSTB, OUTPUT);
    pinMode(SDO, INPUT_PULLUP);
    pinMode(IRQ, INPUT_PULLUP);

    LD_Reset();
    attachInterrupt(IRQ, LD3320::LD_INT, FALLING);
}

void LD3320::LD_WriteReg(unsigned address, unsigned dataout){

    SCS_0;

    // write command, address, data.
    unsigned i, j;
    unsigned k[3]={0x04, address, dataout};
    for (j=0; j<3; j++){
        for (i=0; i<8; i++){
            MYDELAY_NOP;

            if (k[j]&0x80) SDI_1;
            else SDI_0;

            SDCK_0;

            k[j]<<=1; // it is said shift can not used in IRAM_ATTR.
            MYDELAY_NOP;

            SDCK_1;
        }
    }
    MYDELAY_NOP;

    SCS_1;
}

unsigned LD3320::LD_ReadReg(unsigned address){

    SCS_0;

    // write command, address.
    unsigned i, j;
    unsigned k[2]={0x05, address};
    for (j=0; j<2; j++){
        for (i=0; i<8; i++){
            MYDELAY_NOP;

            if (k[j]&0x80) SDI_1;
            else SDI_0;

            SDCK_0;

            k[j]<<=1;
            MYDELAY_NOP;

            SDCK_1;
        }
    }

    // read.
    unsigned datain=0;
    for (i=0; i<8; i++){
        MYDELAY_NOP;
        datain<<=1;

        if (SDO_READ) datain|=1;

        SDCK_0;

        MYDELAY_NOP;

        SDCK_1;
    }
    MYDELAY_NOP;

    SCS_1;

    return datain;
}

/************************************************************************
功能描述：    复位LD模块
入口参数：  none
返 回 值：   none
其他说明：  none
**************************************************************************/
void LD3320::LD_Reset(){

    RSTB_1;
    MYDELAY5;
    RSTB_0;
    MYDELAY_FIXED1;
    RSTB_1;
    MYDELAY5;

    SCS_0;
    MYDELAY_FIXED1;
    SCS_1;
    MYDELAY5;

    nLD_reset=1;
    nLD_Mode=LD_MODE_IDLE;
    nAsrStatus=LD_ASR_IDLE;
}

/************************************************************************
功能描述： LD模块命令初始化
入口参数： none
返 回 值： none
其他说明： 该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_Init_Common(){
    LD_ReadReg(0x06);               // read fifo status
    LD_WriteReg(0x17, 0x35);        // soft reset
    MYDELAY5;

    LD_ReadReg(0x06);
    LD_WriteReg(0x89, 0x03);        // init analog part.
    MYDELAY5;
    LD_WriteReg(0xCF, 0x43);        // init power saving for mp3 or asr.
    MYDELAY5;
    LD_WriteReg(0xCB, 0x02);        // result register4.

    // PLL setting
    LD_WriteReg(0x11, LD_PLL_11);   // clock register1.
    LD_WriteReg(0x1E, 0x00);        // adc init.

    if (nLD_Mode==LD_MODE_MP3){
        LD_WriteReg(0x19, LD_PLL_MP3_19); 
        LD_WriteReg(0x1B, LD_PLL_MP3_1B);   
        LD_WriteReg(0x1D, LD_PLL_MP3_1D);
    }
    else { // LD_MODE_ASR_RUN or LD_MODE_IDLE
        LD_WriteReg(0x19, LD_PLL_ASR_19);
        LD_WriteReg(0x1B, LD_PLL_ASR_1B);
        LD_WriteReg(0x1D, LD_PLL_ASR_1D);
    }
    MYDELAY5;
    LD_WriteReg(0xCD, 0x04);        // allow dsp sleep.
    LD_WriteReg(0x17, 0x4c);        // let dsp sleep.
    MYDELAY5;
    LD_WriteReg(0xB9, 0x00);        // clear the voice-words counter.
    LD_WriteReg(0xCF, 0x4F);        // bring up mp3 or asr.
    LD_WriteReg(0x6F, 0xFF);        // init chip.
}

/************************************************************************
功能描述：    LD模块 ASR功能初始化
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_Init_ASR(){

    nLD_Mode=LD_MODE_ASR_RUN;
    nAsrStatus=LD_ASR_BUSY;

    LD_Init_Common();

    LD_WriteReg(0xBD, 0x00);        // activate asr.
    LD_WriteReg(0x17, 0x48);        // activate dsp.
    MYDELAY5;

    LD_WriteReg(0x3C, 0x80);        // assigning fifo_ext boundary
    LD_WriteReg(0x3E, 0x07);        // assigning fifo_ext boundary
    LD_WriteReg(0x38, 0xff);        // assigning fifo_ext boundary
    LD_WriteReg(0x3A, 0x07);        // assigning fifo_ext boundary

    LD_WriteReg(0x40, 0);           // assigning fifo_ext mcu wm
    LD_WriteReg(0x42, 8);           // assigning fifo_ext mcu wm
    LD_WriteReg(0x44, 0);           // assigning fifo_ext dsp wm
    LD_WriteReg(0x46, 8);           // assigning fifo_ext dsp wm
    MYDELAY1;
}

/************************************************************************
功能描述：    LD模块 MP3功能初始化
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_Init_MP3(){

    nLD_Mode=LD_MODE_MP3;
    nAsrStatus=LD_ASR_BUSY;

    LD_Init_Common();

    LD_WriteReg(0xBD, 0x02);        // activate mp3.
    LD_WriteReg(0x17, 0x48);        // activate dsp.
    MYDELAY5;

    LD_WriteReg(0x85, 0x52);        // internal feedback gain. init it.
    LD_WriteReg(0x8F, 0x00);        // choose lineout, init it.
    LD_WriteReg(0x81, 0x00);        // earphone volume left. set to max.
    LD_WriteReg(0x83, 0x00);        // earphone volume right. set to max.
    LD_WriteReg(0x8E, 0xff);        // speaker volume. set to min.
    LD_WriteReg(0x8D, 0xff);        // internal gain control. init it.
    MYDELAY1;

    LD_WriteReg(0x87, 0xff);        // dac control1. init it for mp3.
    LD_WriteReg(0x89, 0xff);        // dac control2. for mp3 play.
    MYDELAY1;

    LD_WriteReg(0x22, 0x00);        // assigning fifo_data boundary
    LD_WriteReg(0x23, 0x00);        // assigning fifo_data boundary
    LD_WriteReg(0x20, 0xef);        // assigning fifo_data boundary
    LD_WriteReg(0x21, 0x07);        // assigning fifo_data boundary

    LD_WriteReg(0x24, 0x77);        // assigning fifo_data mcu wm
    LD_WriteReg(0x25, 0x03);        // assigning fifo_data mcu wm
    LD_WriteReg(0x26, 0xbb);        // assigning fifo_data dsp wm
    LD_WriteReg(0x27, 0x01);        // assigning fifo_data dsp wm
    MYDELAY1;
}

/************************************************************************
功能描述：    LD模块 MP3 speaker volume
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_AdjustMIX2SPVolume(unsigned val){ // val from 0 to 15.
    LD_WriteReg(0x8E, (((15-val)&0x0f)<<2)|0xC3);
    LD_WriteReg(0x87, 0xFF); // activate volume set; should only bit3 but not ok.
}

/************************************************************************
功能描述：    LD模块 MP3 play
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_Play_MP3(){
    LD_AdjustMIX2SPVolume(AUDIO_VOL); // from 0 to 15. mid volume is 7.
    LD_ReloadMp3Data();

    LD_WriteReg(0xBA, 0x00);        // mp3-end info. clear it.
    LD_WriteReg(0x17, 0x48);        // activate dsp.
    LD_WriteReg(0x33, 0x01);        // start to play.
    LD_WriteReg(0x29, 0x04);        // enable fifo-int.
    LD_WriteReg(0x02, 0x01);        // enable fifo-data int.
    LD_WriteReg(0x85, 0x5A);        // internal feedback gain. play mp3 change gain to 30kohm.
}

/************************************************************************
功能描述：    LD模块 MP3 data transmission
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
void LD3320::LD_ReloadMp3Data(){ // notice byte and int and little-endian and signed.
    unsigned mp3data;
    while (!(LD_ReadReg(0x06)&MASK_FIFO_STATUS_AFULL) && LD3320MP3::Read(mp3data))
        LD_WriteReg(0x01, (char)(*((unsigned char*)&mp3data))); // write fifo_data.

    // the original procedure is simple: read specific data bytes from flash, a byte each time -> check ld3320 fifo_data not full ->
    // write this byte -> until either fifo is full(and to do it later) or entire mp3 file is done.
}

/************************************************************************
功能描述：    LD模块 MP3 int isr
入口参数：  none
返 回 值：   none
其他说明：  该函数为出厂配置，一般不需要修改；
           有兴趣的客户可对照开发手册根据需要自行修改。
**************************************************************************/
#if !(MP3_IN_ISR)
void LD3320::ProcessIntMP3(){
    if (LD_ReadReg(0xBA)&CAUSE_MP3_SONG_END){ // check whether MP3-end which sensed by chip.
        LD_WriteReg(0x2B, 0x00); // clear the int flag.
        LD_WriteReg(0xBA, 0x00); // clear it.
        LD_WriteReg(0xBC, 0x00); // 01? to inform the mp3 ended up.
        LD_WriteReg(0x08, 0x01); // clear the fifo_data, set it.
        LD_WriteReg(0x08, 0x00); // clear the fifo_data, do it.
        LD_WriteReg(0x33, 0x00); // done the play.

        LD3320MP3::Close();
        LD_Reset();
        nAsrStatus=LD_ASR_NONE;
        return;
    }
    else if (LD3320MP3::isDone() || !LD3320MP3::isReady()){
        LD_WriteReg(0xBC, 0x01); // to inform the mp3 ended up, which sensed by us.
        LD_WriteReg(0x29, 0x10); // enable sync-int. for what?

        LD3320MP3::Close();
        LD_Reset();
        nAsrStatus=LD_ASR_NONE;
        return;
    }
    LD_ReloadMp3Data();
    LD_WriteReg(0x29, 0x04); // enable fifo-int.
    LD_WriteReg(0x02, 0x01); // enable fifo-data int.
    nAsrStatus=LD_ASR_IDLE; // and waits for interrupt.
}
#endif // !(MP3_IN_ISR)

/************************************************************************
功能描述：  检测LD模块是否空闲
入口参数： none
返 回 值：  flag：1-> 空闲
其他说明： none
**************************************************************************/
unsigned LD3320::LD_Check_ASRBusyFlag_b2(){ // suggest to reset if always busy.
    for (int i=0; i<100; i++){
        if (LD_ReadReg(0xB2)==0x21) return 1;
        MYDELAY20;
    }
    nAsrStatus=LD_ASR_ERROR;

    ld_fptr=0;

    return 0;
}

/************************************************************************
功能描述：   运行ASR
入口参数： none
返 回 值：  1：启动成功
其他说明： none
**************************************************************************/
unsigned LD3320::LD_AsrRun(){
    LD_WriteReg(0x35, MIC_VOL);     // set vol.
    LD_WriteReg(0x1C, 0x09);        // activate adc, now use default.
    LD_WriteReg(0xBD, 0x20);        // activate mp3 or asr, now use default.
    LD_WriteReg(0x08, 0x01);        // clear fifo_data.
    MYDELAY1;
    LD_WriteReg(0x08, 0x00);        // issue clear fifo_data.
    MYDELAY5;

    if (LD_Check_ASRBusyFlag_b2()){
        LD_WriteReg(0xB2, 0xff);    // asr, dsp, busy state.
        LD_WriteReg(0x37, 0x06);    // ask dsp to start to voice recog.
        MYDELAY20;
        LD_WriteReg(0x1C, 0x0b);    // activate adc, now mic channel enabled. pin 9, 10, 12.
        LD_WriteReg(0x29, 0x10);    // enable sync-int.
        LD_WriteReg(0xBD, 0x00);    // activate asr.
        return 1;
    }

    return 0; // failed.
}

/************************************************************************
功能描述：  change table
入口参数： table
返 回 值：  none
其他说明： none
**************************************************************************/
void LD3320::LD_ChangeTable(void *t){ // t, a table pointer.
    var_tbl=(char*(*)[])t;
    nLD_change_table=1;
}

/************************************************************************
功能描述：  get table
入口参数： none
返 回 值：  a table pointer
其他说明： none
**************************************************************************/
void* LD3320::LD_GetTable(){ // return the current table pointer.
    return var_tbl;
}

/************************************************************************
功能描述： 向LD模块添加关键词
入口参数： none
返 回 值： flag：1->添加成功
其他说明： 用户修改.
           1、根据如下格式添加拼音关键词，同时注意修改sRecog 和pCode 数组的长度
           和对应变了k的循环置。拼音串和识别码是一一对应的。
           2、开发者可以学习"语音识别芯片LD3320高阶秘籍.pdf"中
           关于垃圾词语吸收错误的用法，来提供识别效果。
**************************************************************************/
unsigned LD3320::LD_AsrAddFixed(){

    for (int i=0; *((*var_tbl)[i]); i++){

        if (!LD_Check_ASRBusyFlag_b2()) return 0;

        LD_WriteReg(0xc1, i);       // the id from 00 to ff.
        LD_WriteReg(0xc3, 0x00);    // ready to add sentence.
        LD_WriteReg(0x08, 0x04);    // clear the fifo_ext, set it.
        MYDELAY1;
        LD_WriteReg(0x08, 0x00);    // clear the fifo_ext, do it.
        MYDELAY1;

        int j; unsigned k;
        for (j=0; k=(*var_tbl)[i][j]; j++) LD_WriteReg(0x5, k); // fifo_ext data port.

        LD_WriteReg(0xb9, j);       // terminated by assigning the length of sentence.
        LD_WriteReg(0xb2, 0xff);    // asr, dsp, busy state. clear it for later check.
        LD_WriteReg(0x37, 0x04);    // notify the dsp for adding a sentence.
        MYDELAY_NOP;
    }
    return 1;
}

/************************************************************************
功能描述：   运行ASR识别流程
入口参数： none
返 回 值：  asrflag：1->启动成功， 0—>启动失败
其他说明： 识别顺序如下:
            1、RunASR()函数实现了一次完整的ASR语音识别流程
            2、LD_AsrStart() 函数实现了ASR初始化
            3、LD_AsrAddFixed() 函数实现了添加关键词语到LD3320芯片中
            4、LD_AsrRun() 函数启动了一次ASR语音识别流程          
            任何一次ASR识别流程，都需要按照这个顺序，从初始化开始
**************************************************************************/
unsigned LD3320::RunASR(p1fptr a){ // must LD_Reset() if failed. parameter: callback function.
    ld_fptr=a;

#if 0
    nLD_reset=0;
    nLD_change_table=0;
    nLD_Mode=LD_MODE_ASR_RUN;
    nAsrStatus=LD_ASR_BUSY;

    LD_Init_ASR();
    if (LD_AsrAddFixed()) return LD_AsrRun();
    return 0; // failed.
#else

    static unsigned try1, try2, try3;
    if (nLD_reset){try1=1, try2=1, try3=1, nLD_reset=0;}
    //// if (nLD_change_table){nLD_change_table=0; try3=1;} // this is not ok, new table high errors.
    if (nLD_change_table){nLD_change_table=0; try1=1, try2=1, try3=1;} ////

    if (try1){
        try1=0;

        nLD_Mode=LD_MODE_ASR_RUN;
        nAsrStatus=LD_ASR_BUSY;

        LD_ReadReg(0x06);               // read fifo status
        LD_WriteReg(0x17, 0x35);        // soft reset
        MYDELAY5;

        LD_ReadReg(0x06);
        LD_WriteReg(0x89, 0x03);        // init analog part.
        MYDELAY5;
        LD_WriteReg(0xCF, 0x43);        // init power saving for mp3 or asr.
        MYDELAY5;
        LD_WriteReg(0xCB, 0x02);        // result register4.

        // PLL setting
        LD_WriteReg(0x11, LD_PLL_11);   // clock register1.
        LD_WriteReg(0x1E, 0x00);        // adc init.
        // LD_MODE_ASR_RUN or LD_MODE_IDLE
        LD_WriteReg(0x19, LD_PLL_ASR_19);
        LD_WriteReg(0x1B, LD_PLL_ASR_1B);
        LD_WriteReg(0x1D, LD_PLL_ASR_1D);

        MYDELAY5;
        LD_WriteReg(0xCD, 0x04);        // allow dsp sleep.
        LD_WriteReg(0x17, 0x4c);        // let dsp sleep.
        MYDELAY5;
        LD_WriteReg(0xB9, 0x00);        // clear the voice-words counter.
        LD_WriteReg(0xCF, 0x4F);        // bring up mp3 or asr.
        LD_WriteReg(0x6F, 0xFF);        // init chip.

        LD_WriteReg(0xBD, 0x00);        // activate asr.
        LD_WriteReg(0x17, 0x48);        // activate dsp.
        MYDELAY5;
    }

    if (try2){
        try2=0;

        LD_WriteReg(0x3C, 0x80);        // assigning fifo_ext boundary
        LD_WriteReg(0x3E, 0x07);        // assigning fifo_ext boundary
        LD_WriteReg(0x38, 0xff);        // assigning fifo_ext boundary
        LD_WriteReg(0x3A, 0x07);        // assigning fifo_ext boundary

        LD_WriteReg(0x40, 0);           // assigning fifo_ext mcu wm
        LD_WriteReg(0x42, 8);           // assigning fifo_ext mcu wm
        LD_WriteReg(0x44, 0);           // assigning fifo_ext dsp wm
        LD_WriteReg(0x46, 8);           // assigning fifo_ext dsp wm
        MYDELAY1;
    }

    if (try3){
        try3=0;

        if (!LD_AsrAddFixed()) return 0;
    }

    nLD_Mode=LD_MODE_ASR_RUN;
    nAsrStatus=LD_ASR_BUSY;
    return LD_AsrRun();

#endif
}

/************************************************************************
功能描述：   运行MP3 play
入口参数： none
返 回 值：  none
其他说明： none
**************************************************************************/
unsigned LD3320::RunMP3(){ // must LD_Reset().
    // do not repeat call during a play, otherwise some file may be reopen or orphan and cause stack overflow.

    if (!LD3320MP3::isReady()) return 0;

    LD_Reset();

    LD_Init_MP3();
    LD_Play_MP3();
    return 1;
}

/************************************************************************
功能描述：   stop ASR
入口参数： none
返 回 值：  none
其他说明： none
**************************************************************************/
void LD3320::StopASR(){ // force asr to stop.
    LD_WriteReg(0xBC, 0x07);            // stop asr sampling.
    MYDELAY5;
    LD_WriteReg(0xBC, 0x08);            // stop asr calculation.
    MYDELAY5;

    LD_WriteReg(0x2B, 0x00);            // clear the int flag.
    nLD_reset=1;                        // later to do whole procedure rather than simplified.

    nLD_Mode=LD_MODE_IDLE;
    nAsrStatus=LD_ASR_IDLE;
}

/************************************************************************
功能描述：   获取识别结果
入口参数： none
返 回 值：  LD_ReadReg(0xc5 )；  读取内部寄存器返回识别码。
其他说明： none
**************************************************************************/
unsigned LD3320::LD_GetResult(){return LD_ReadReg(0xc5);} // the result register1.

/************************************************************************
功能描述：   asr callback
入口参数： none
返 回 值：  none
其他说明： none
**************************************************************************/
void LD3320::LD_Callback(unsigned id){ld_fptr(id); ld_fptr=0;}

/************************************************************************
功能描述：   interrupt
入口参数： none
返 回 值：  none
其他说明： none
**************************************************************************/
void LD3320::LD_INT(){

    unsigned int_flag=LD_ReadReg(0x2B); // get the int flags.

    unsigned ucHighInt=LD_ReadReg(0x29); // read the int enable register1.
    unsigned ucLowInt=LD_ReadReg(0x02); // read the int enable register2.

    LD_WriteReg(0x29, 0);               // disable int.
    LD_WriteReg(0x02, 0);               // disable fifo int.

    if (int_flag&0x10){                 // int from asr.
        if ((LD_ReadReg(0xb2)==0x21) && (LD_ReadReg(0xbf)==0x35)){
            unsigned cnt=LD_ReadReg(0xba);
            if (cnt && cnt<5){          // 4 >= the match counts >= 1.
                nAsrStatus=LD_ASR_FOUNDOK;
                if (ld_fptr) LD_Callback(LD_GetResult());
            }
            else nAsrStatus=LD_ASR_FOUNDZERO;
        }
        else nAsrStatus=LD_ASR_EXCEPTION;
    }
    else if (int_flag&0x08){            // internal error.
        nAsrStatus=LD_ASR_INTERR;
    }
    else { // mp3 interrupt. // (nLD_Mode == LD_MODE_MP3)
        nAsrStatus=LD_ASR_MP3;          // mp3 int.

        #if (MP3_IN_ISR) // else goto use ProcessIntMP3();
        if (LD_ReadReg(0xBA)&CAUSE_MP3_SONG_END){ // check whether MP3-end which sensed by chip.
            LD_WriteReg(0x2B, 0x00); // clear the int flag.
            LD_WriteReg(0xBA, 0x00); // clear it.
            LD_WriteReg(0xBC, 0x00); // 01? to inform the mp3 ended up.
            LD_WriteReg(0x08, 0x01); // clear the fifo_data, set it.
            LD_WriteReg(0x08, 0x00); // clear the fifo_data, do it.
            LD_WriteReg(0x33, 0x00); // done the play.

            LD3320MP3::Close();
            ucHighInt=ucLowInt=0;
            LD_Reset();
            nAsrStatus=LD_ASR_NONE;
        }
        else if (LD3320MP3::isDone() || !LD3320MP3::isReady()){
            LD_WriteReg(0xBC, 0x01); // to inform the mp3 ended up, which sensed by us.
            LD_WriteReg(0x29, 0x10); // enable sync-int. for what?

            LD3320MP3::Close();
            ucHighInt=ucLowInt=0;
            LD_Reset();
            nAsrStatus=LD_ASR_NONE;
        }
        else LD_ReloadMp3Data();
        #else // !(MP3_IN_ISR)
        ucHighInt=ucLowInt=0;
        #endif // (MP3_IN_ISR)

        LD_WriteReg(0x2b, 0);           // clear the int flags.
        LD_WriteReg(0x29, ucHighInt);   // restore the int enable register1.
        LD_WriteReg(0x02, ucLowInt);    // restore the int enable register2.

        return;
    }

    LD_WriteReg(0x2b, 0);               // clear the int flags.
    LD_WriteReg(0x1C, 0);               // disable adc.
    LD_WriteReg(0xBC, 0x07);            // force to stop asr.
    LD_WriteReg(0xBA, 0);               // clear the match count.
    LD_WriteReg(0x08, 0x01);            // clear fifo_data.
    MYDELAY1;
    LD_WriteReg(0x08, 0x00);            // issue clear fifo_data.
}

/************************************************************************
功能描述：   main loop
入口参数： none
返 回 值：  none
其他说明： none
**************************************************************************/
void LD3320::LD_LoopEx(){ // regularly run the asr or play an mp3 file.
    // another way is to use RunASR(fptr), fptr(id) is callbacked once recognized success.

    switch (nAsrStatus){
        case LD_ASR_IDLE:
        case LD_ASR_BUSY: break;
        case LD_ASR_NONE: // regularly run the asr or play an mp3 file.
            if (LD3320::RunMP3()) break; // use LD3320MP3 tmp("sample.mp3") first to trigger.
            else if (!LD3320::RunASR()){
                // Serial.println("ASR fault and reset and retry.");
                LD3320::LD_Reset();
                nAsrStatus=LD_ASR_NONE;
                delay(50);
            }
            break;
        case LD_ASR_FOUNDOK: // asr success.
            {
                unsigned b=LD3320::LD_GetResult();
                // ((void(*)(unsigned))((*(voice_tables->nxtT()))[b]))(b); // call the target.
            }
            nAsrStatus=LD_ASR_NONE;
            break;
        case LD_ASR_FOUNDZERO:
            // Serial.println("ASR done but no result.");
            nAsrStatus=LD_ASR_NONE;
            break;
        case LD_ASR_MP3: // set by mp3 intr.
            #if !(MP3_IN_ISR)
                LD3320::ProcessIntMP3();
            #endif
            break;
        default: // LD_ASR_ERROR from busy; LD_ASR_EXCEPTION or LD_ASR_INTERR from intr.
            // Serial.printf("Unexpected error (%u).\r\n", nAsrStatus);
            nAsrStatus=LD_ASR_NONE;
        break;
    }
}
