// rearranged LD3320 sample code, for tests on both asr and mp3 playing by ken woo 20220614.v.0.2.
// 20220618 v.0.3, support esp-m3 which disabled tx/rx. and with some structural modifies.


#ifndef __LD3320_H
#define __LD3320_H
#include <ESP8266WiFi.h>
#include<LittleFS.h>


/////////////////// user modify here.
#define MP3_IN_ISR      1 // if in isr, the isr has to access flash then may cause problem,
                        // instead, only set flag for later main loop to access, however might affect playing.
#define ESP8266_SPEEDUP 1 // use esp8266 faster io.

#define USE_ESP_M3      0

#define InTheIram       IRAM_ATTR

#define AUDIO_VOL       7 // 7 is mid.

#if !(USE_ESP_M3)
// modified for esp8266, wemos d1 mini.
    #define SDCK            5 // SPI时钟信号, D1.
    #define SDO             4 // SPI slave data output, MI, 数据输出, D2.
    #define SDI             14 // SPI, MO, 数据输入, D5.
    #define SCS             12 // 芯片片选信号, D6.
    #define RSTB            2 // 复位端口, D4.
    #define IRQ             13 // 中断引脚, D7.
#else // USE_ESP_M3
// modified for esp8285, esp-m3-1m. should use #define MYDEBUG 0 in main().
    #define SDCK            12 // SPI时钟信号
    #define SDO             3 // SPI slave data output, MI, 数据输出
    #define SDI             1 // SPI, MO, 数据输入
    #define SCS             0 // 芯片片选信号
    #define RSTB            14 // 复位端口
    #define IRQ             4 // 中断引脚
#endif




#define MYDELAY1        delayMicroseconds(1)//delay(1)
#define MYDELAY5        delayMicroseconds(10)//delay(5)
#define MYDELAY20       delayMicroseconds(15)//delay(20)
#define MYDELAY_NOP     delayMicroseconds(1)
#define MYDELAY_FIXED1  delay(1)


#if ESP8266_SPEEDUP // for speed up spi timing for esp8266.
    extern volatile unsigned &PIN_OUT_SET;
    extern volatile unsigned &PIN_OUT_CLEAR;
    extern volatile unsigned &PIN_IN;

    #define IO00    0x00001
    #define IO01    0x00002
    #define IO02    0x00004
    #define IO03    0x00008
    #define IO04    0x00010
    #define IO05    0x00020
    #define IO12    0x01000
    #define IO13    0x02000
    #define IO14    0x04000
    #define IO15    0x08000

    #if !(USE_ESP_M3)
        #define SDCK_1 (PIN_OUT_SET=IO05)
        #define SDCK_0 (PIN_OUT_CLEAR=IO05)

        #define SDI_1 (PIN_OUT_SET=IO14)
        #define SDI_0 (PIN_OUT_CLEAR=IO14)

        #define SCS_1 (PIN_OUT_SET=IO12)
        #define SCS_0 (PIN_OUT_CLEAR=IO12)

        #define RSTB_1 (PIN_OUT_SET=IO02)
        #define RSTB_0 (PIN_OUT_CLEAR=IO02)

        #define SDO_READ (PIN_IN&IO04)
    #else // (USE_ESP_M3)
        #define SDCK_1 (PIN_OUT_SET=IO12)
        #define SDCK_0 (PIN_OUT_CLEAR=IO12)

        #define SDI_1 (PIN_OUT_SET=IO01)
        #define SDI_0 (PIN_OUT_CLEAR=IO01)

        #define SCS_1 (PIN_OUT_SET=IO00)
        #define SCS_0 (PIN_OUT_CLEAR=IO00)

        #define RSTB_1 (PIN_OUT_SET=IO14)
        #define RSTB_0 (PIN_OUT_CLEAR=IO14)

        #define SDO_READ (PIN_IN&IO03)
    #endif

#else // ordinary arduino io access.
    #define SDCK_1 digitalWrite(SDCK, HIGH)
    #define SDCK_0 digitalWrite(SDCK, LOW)

    #define SDI_1 digitalWrite(SDI, HIGH)
    #define SDI_0 digitalWrite(SDI, LOW)

    #define SCS_1 digitalWrite(SCS, HIGH)
    #define SCS_0 digitalWrite(SCS, LOW)

    #define RSTB_1 digitalWrite(RSTB, HIGH)
    #define RSTB_0 digitalWrite(RSTB, LOW)

    #define SDO_READ (digitalRead(SDO))
#endif


//  以下三个状态定义用来记录程序是在运行ASR识别还是在运行MP3播放
#define LD_MODE_IDLE      0x00
#define LD_MODE_ASR_RUN   0x08
#define LD_MODE_MP3       0x40


// for playing mp3
#define MASK_FIFO_STATUS_AFULL      0x08 // fifo_data full if bitset.
#define CAUSE_MP3_SONG_END          0x20 // chip saw the mp3-file-end.


// 以下状态定义用来记录程序是在运行ASR识别过程中的哪个状态
#define LD_ASR_IDLE       0x00  //  表示没有在作ASR识别
#define LD_ASR_NONE       0x01  //  available and going do asr or play mp3.
#define LD_ASR_BUSY       0x02  //  表示LD3320正在作ASR识别中
#define LD_ASR_FOUNDOK    0x04  //  表示一次识别流程结束后，有一个识别结果
#define LD_ASR_FOUNDZERO  0x08  //  表示一次识别流程结束后，没有识别结果
#define LD_ASR_ERROR      0x10  //  表示一次识别流程中LD3320芯片内部出现不正确的状态
#define LD_ASR_EXCEPTION  0x11  //  exception
#define LD_ASR_INTERR     0x12  //  interrupt error
#define LD_ASR_MP3        0x14  //  mp3


#define CLK_IN          24  /* 用户注意修改输入的晶振时钟大小 */
#define LD_PLL_11       (unsigned char)((CLK_IN/2.0)-1)
#define LD_PLL_MP3_19   0x0f
#define LD_PLL_MP3_1B   0x18
#define LD_PLL_MP3_1D   (unsigned char)(((90.0*((LD_PLL_11)+1))/(CLK_IN))-1)

#define LD_PLL_ASR_19   (unsigned char)(CLK_IN*32.0/(LD_PLL_11+1) - 0.51)
#define LD_PLL_ASR_1B   0x48
#define LD_PLL_ASR_1D   0x1f


#define MIC_VOL 0x43   //咪头增益


typedef class LD3320{ // all static.

    typedef void (*fptr)();                     // function pointer without parameter.
    typedef void (*p1fptr)(unsigned);           // function pointer with 1 parameter.

    // vars
    InTheIram static unsigned nLD_Mode;         // the mode state.
    InTheIram static unsigned nLD_reset;        // demand to reset/go for whole procedure.
    InTheIram static char *(*var_tbl)[];        // the table pointer used for loading the table.
    InTheIram static unsigned nLD_change_table; // demand to change table.

    // callback func
    IRAM_ATTR static p1fptr ld_fptr;            // for asr. callbacked if set, and used in one-shot.
    IRAM_ATTR static void LD_Callback(unsigned id); // for asr. called only if ld_fptr is set, and then set it to 0.

    // basics
    IRAM_ATTR static void LD_WriteReg(unsigned address, unsigned dataout);
    IRAM_ATTR static unsigned LD_ReadReg(unsigned address);
    InTheIram static void LD_Init_Common();
    InTheIram static void LD_Init_ASR();
    InTheIram static unsigned LD_AsrRun();
    InTheIram static unsigned LD_AsrAddFixed();
    InTheIram static unsigned LD_Check_ASRBusyFlag_b2();

    // mp3s
    InTheIram static void LD_Init_MP3();
    InTheIram static void LD_Play_MP3();
    IRAM_ATTR static void LD_ReloadMp3Data();

        public:

    // basics
    static void LD_Init();
    IRAM_ATTR static void LD_Reset();
    IRAM_ATTR static void LD_INT();
    IRAM_ATTR static unsigned LD_GetResult();

    // asrs
    static unsigned RunASR(p1fptr a=0);
    static void StopASR();

    // mp3s
    #if !(MP3_IN_ISR)
        IRAM_ATTR static void ProcessIntMP3();
    #endif
    static unsigned RunMP3();
    static void LD_AdjustMIX2SPVolume(unsigned val);

    // helper funcs
    static void LD_ChangeTable(void *t); // t, a table pointer.
    static void* LD_GetTable(); // return the current table pointer.
    static void LD_LoopEx(); // main loop example.

} LD3320;

extern unsigned nAsrStatus;


typedef class LD3320MP3{ // note that since it is byte-manipulation, so it can not in IRAM_ATTR(for Read).
    // reasons that, over here is single threaded. only one mp3 file is played at a time.
    // close it before proceeding another file is ok.
    // and the filesystem we use uses buffer for any number of bytes we request which is ok too, i.e., so 1.
    // so we can simply use a static member function to get data and must explicitly close before for another file.
    IRAM_ATTR static File fh;
    IRAM_ATTR static unsigned flock;
    InTheIram static const char *var_mp3_filename;
        public:
    IRAM_ATTR static unsigned Read(unsigned &r_byte){ // since function calls are in flash and inevitable.
        // return true if success, false if file-end or errors.
        int i=0;
        unsigned j;
        if (fh || flock || (fh=LittleFS.open(var_mp3_filename, "r"))){
            if (flock!=2) i=fh.read((unsigned char*)(&j), 1);
            if (i!=1){
                if (flock!=2){
                    var_mp3_filename=0;
                    flock=2;
                    fh.close();
                    fh=File(0);
                    LittleFS.end();
                }
                return 0;
            }
            flock=1;
            r_byte=j;
            return 1;
        }
        return 0;
    };
    LD3320MP3(){;};
    LD3320MP3(const char *a){if (!a || var_mp3_filename) return; LittleFS.begin(); var_mp3_filename=a;};
    IRAM_ATTR static unsigned isDone(){return (flock==2);};
    IRAM_ATTR static unsigned isReady(){
        if (var_mp3_filename){
            if (LittleFS.exists(var_mp3_filename)) return 1;

            // may affect status but might more safe.
            var_mp3_filename=0;
            flock=0;
            if (fh){fh.close(); fh=File(0);}
            LittleFS.end();
        }
        return 0;
    };
    IRAM_ATTR static void Close(){
        if (var_mp3_filename){
            if (fh){fh.close(); fh=File(0);}
            var_mp3_filename=0;
            LittleFS.end();
        }
        flock=0;
    };
    InTheIram static unsigned isOpened(){return (var_mp3_filename || fh || flock);}; // must check closed before playing another file.
} LD3320MP3;


#endif // __LD3320_H
